# defmodule Remedy.Voice do
#   @moduledoc """
#   Interface for playing audio through Discord's voice channels.

#   # Using Discord Voice Channels
#   To play sound in Discord with Remedy, you'll need `ffmpeg` to be installed.
#   If you don't have the executable `ffmpeg` in the path, the absolute path may
#   be configured through config keys `:remedy, :ffmpeg`. If you don't want to use
#   ffmpeg, read on to the next section.

#   A bot may be connected to at most one voice channel per guild. For this reason,
#   most of the functions in this module take a guild id, and the resulting action
#   will be performed in the given guild's voice channel that the bot is connected to.

#   The primary Discord gateway responsible for all text based communication relies on
#   one WSState connection per shard, where small bots typically only have one shard.
#   The Discord voice gateways work by establishing a WSState connection per guild/channel.
#   After some handshaking on this connection, audio data can be sent over UDP/RTP. Behind
#   the scenes the voice WSState connections are implemented nearly the same way the main
#   shard WSState connections are, and require no developer intervention.

#   ## Voice Without FFmpeg
#   If you wish to BYOE (Bring Your Own Encoder), there are a few options.
#     - Use `:raw` as `type` for `Remedy.Voice.play/4`
#       - Provide the complete list of opus frames as the input
#     - Use `:raw_s` as `type` for `Remedy.Voice.play/4`
#       - Provide a stateful enumerable of opus frames as input (think GenServer wrapped in `Stream.unfold/2`)
#     - Use lower level functions to send opus frames at your leisure
#       - Send packets on your own time using `Remedy.Voice.send_frames/2`
#   """

#   alias Porcelain.Process, as: Proc
#   alias Remedy.API

#   alias Remedy.Voice.{Audio, Session}
#   alias Remedy.Voice.Supervisor, as: VoiceSupervisor
#   use Remedy.Schema
#   require Logger

#   use GenServer

#   @doc false
#   def start_link(_args) do
#     GenServer.start_link(__MODULE__, %{}, name: VoiceStateMap)
#   end

#   @doc false
#   def init(args) do
#     {:ok, args}
#   end

#   @doc false
#   def update_voice(guild_id, args \\ []) do
#     GenServer.call(VoiceStateMap, {:update, guild_id, args})
#   end

#   @doc false
#   def get_voice(guild_id) do
#     GenServer.call(VoiceStateMap, {:get, guild_id})
#   end

#   @doc false
#   def remove_voice(guild_id) do
#     GenServer.call(VoiceStateMap, {:remove, guild_id})
#   end

#   @doc """
#   Joins or moves the bot to a voice channel.

#   This function is equivalent to `Remedy.API.update_voice_state/4`.
#   """
#   @spec join_channel(Guild.id(), Channel.id(), boolean, boolean) :: no_return | :ok
#   def join_channel(guild_id, channel_id, self_mute \\ false, self_deaf \\ false) do
#     API.update_voice_state(guild_id, channel_id, self_mute, self_deaf)
#   end

#   @doc """
#   Leaves the voice channel of the given guild id.

#   This function is equivalent to calling `Remedy.API.update_voice_state(guild_id, nil)`.
#   """
#   @spec leave_channel(Guild.id()) :: no_return | :ok
#   def leave_channel(guild_id) do
#     API.update_voice_state(guild_id, nil)
#   end

#   @doc """
#   Plays sound in the voice channel the bot is in.

#   The bot must be connected to a voice channel in the guild specified.

#   ## Parameters
#     - `guild_id` - ID of guild whose voice channel the sound will be played in.
#     - `input` - Audio to be played. Type of `input` determined by `type` parameter.
#     - `type` - Type of input (defaults to `:url`).
#       - `:url` Input will be [any url that `ffmpeg` can read](https://www.ffmpeg.org/ffmpeg-protocols.html).
#       - `:pipe` Input will be data that is piped to stdin of `ffmpeg`.
#       - `:ytdl` Input will be url for `youtube-dl`, which gets automatically piped to `ffmpeg`.
#       - `:stream` Input will be livestream url for `streamlink`, which gets automatically piped to `ffmpeg`.
#       - `:raw` Input will be an enumarable of raw opus frames. This bypasses `ffmpeg` and all options.
#       - `:raw_s` Same as `:raw` but input must be stateful, i.e. calling `Enum.take/2` on input is not idempotent.
#     - `options` - See options section below.

#   Returns `{:error, reason}` if unable to play or a sound is playing, else `:ok`.

#   ## Options
#     - `:start_pos` (string) - The start position of the audio to be played. Defaults to beginning.
#     - `:duration` (string) - The duration to of the audio to be played . Defaults to entire duration.
#     - `:realtime` (boolean) - Make ffmpeg process the input in realtime instead of as fast as possible. Defaults to true.
#     - `:volume` (number) - The output volume of the audio. Default volume is 1.0.
#     - `:filter` (string) - Filter(s) to be applied to the audio. No filters applied by default.

#     The values of `:start_pos` and `:duration` can be [any time duration that ffmpeg can read](https://ffmpeg.org/ffmpeg-utils.html#Time-duration).
#     The `:filter` can be used multiple times in a single call (see examples).
#     The values of `:filter` can be [any audio filters that ffmpeg can read](https://ffmpeg.org/ffmpeg-filters.html#Audio-Filters).
#     Filters will be applied in order and can be as complex as you want. The world is your oyster!

#     Note that using the `:volume` option is shortcut for the "volume" filter, and will be added to the end of the filter chain, acting as a master volume.
#     Volume values between `0.0` and `1.0` act as standard oparating range where `0` is off and `1` is max.
#     Values greater than `1.0` will add saturation and distortion to the audio.
#     Negative values act the same as their position but reverse the polarity of the waveform.

#     Having all the ffmpeg audio filters available is *extremely powerful* so it may be worth learning some of them for your use cases.
#     If you use any filters to *increase* the playback speed of your audio, it's recommended to set the `:realtime` option to `false`
#     because realtime processing is relative to the original playback speed.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "~/music/FavoriteSong.mp3", :url)

#   iex> Remedy.Voice.play(123456789, "~/music/NotFavoriteButStillGoodSong.mp3", :url, volume: 0.5)

#   iex> Remedy.Voice.play(123456789, "~/music/ThisWillBeHeavilyDistorted.mp3", :url, volume: 1000)
#   ```
#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> raw_data = File.read!("~/music/sound_effect.wav")

#   iex> Remedy.Voice.play(123456789, raw_data, :pipe)
#   ```
#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "https://www.youtube.com/watch?v=b4RJ-QGOtw4", :ytdl,
#   ...>   realtime: true, start_pos: "0:17", duration: "30")

#   iex> Remedy.Voice.play(123456789, "https://www.youtube.com/watch?v=0ngcL_5ekXo", :ytdl,
#   ...>   filter: "lowpass=f=1200", filter: "highpass=f=300", filter: "asetrate=44100*0.5")
#   ```
#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "https://www.twitch.tv/pestily", :stream)

#   iex> Remedy.Voice.play(123456789, "https://youtu.be/LN4r-K8ZP5Q", :stream)
#   ```
#   """
#   @spec play(
#           Guild.id(),
#           String.t() | binary() | iodata() | Enum.t(),
#           :url | :pipe | :ytdl | :stream | :raw | :raw_s,
#           keyword()
#         ) ::
#           :ok | {:error, String.t()}
#   def play(guild_id, input, type \\ :url, options \\ []) do
#     voice = get_voice(guild_id)

#     cond do
#       not VoiceState.ready_for_rtp?(voice) ->
#         {:error, "Must be connected to voice channel to play audio."}

#       VoiceState.playing?(voice) ->
#         {:error, "Audio already playing in voice channel."}

#       true ->
#         unless is_nil(voice.ffmpeg_proc), do: Proc.stop(voice.ffmpeg_proc)
#         set_speaking(voice, true)

#         {ffmpeg_proc, raw_audio, raw_stateful} =
#           case type do
#             :raw -> {nil, input, false}
#             :raw_s -> {nil, input, true}
#             _ffmpeg -> {Audio.spawn_ffmpeg(input, type, options), nil, false}
#           end

#         voice =
#           update_voice(guild_id,
#             ffmpeg_proc: ffmpeg_proc,
#             raw_audio: raw_audio,
#             raw_stateful: raw_stateful
#           )

#         {:ok, pid} = Task.start(fn -> Audio.init_player(voice) end)
#         update_voice(guild_id, player_pid: pid)
#         :ok
#     end
#   end

#   @doc """
#   Stops the current sound being played in a voice channel.

#   The bot must be connected to a voice channel in the guild specified.

#   ## Parameters
#     - `guild_id` - ID of guild whose voice channel the sound will be stopped in.

#   Returns `{:error, reason}` if unable to stop or no sound is playing, else `:ok`.

#   If a sound has finished playing, this function does not need to be called to start
#   playing another sound.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "http://brandthill.com/files/weird_dubstep_noises.mp3")

#   iex> Remedy.Voice.stop(123456789)
#   ```
#   """
#   @spec stop(Guild.id()) :: :ok | {:error, String.t()}
#   def stop(guild_id) do
#     voice = get_voice(guild_id)

#     cond do
#       not VoiceState.ready_for_rtp?(voice) ->
#         {:error, "Must be connected to voice channel to stop audio."}

#       not VoiceState.playing?(voice) ->
#         {:error, "Audio must be playing to stop."}

#       true ->
#         set_speaking(voice, false)
#         Process.exit(voice.player_pid, :stop)
#         unless is_nil(voice.ffmpeg_proc), do: Proc.stop(voice.ffmpeg_proc)
#         :ok
#     end
#   end

#   @doc """
#   Pauses the current sound being played in a voice channel.

#   The bot must be connected to a voice channel in the guild specified.

#   ## Parameters
#     - `guild_id` - ID of guild whose voice channel the sound will be paused in.

#   Returns `{:error, reason}` if unable to pause or no sound is playing, else `:ok`.

#   This function is similar to `stop/1`, except that the sound may be
#   resumed after being paused.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "~/files/twelve_hour_loop_of_waterfall_sounds.mp3")

#   iex> Remedy.Voice.pause(123456789)
#   ```
#   """
#   @spec pause(Guild.id()) :: :ok | {:error, String.t()}
#   def pause(guild_id) do
#     voice = get_voice(guild_id)

#     cond do
#       not VoiceState.ready_for_rtp?(voice) ->
#         {:error, "Must be connected to voice channel to pause audio."}

#       not VoiceState.playing?(voice) ->
#         {:error, "Audio must be playing to pause."}

#       true ->
#         set_speaking(voice, false)
#         Process.exit(voice.player_pid, :pause)
#         :ok
#     end
#   end

#   @doc """
#   Resumes playing the current paused sound in a voice channel.

#   The bot must be connected to a voice channel in the guild specified.

#   ## Parameters
#     - `guild_id` - ID of guild whose voice channel the sound will be resumed in.

#   Returns `{:error, reason}` if unable to resume or no sound has been paused, otherwise returns `:ok`.

#   This function is used to resume a sound that had previously been paused.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "~/stuff/Toto - Africa (Bass Boosted)")

#   iex> Remedy.Voice.pause(123456789)

#   iex> Remedy.Voice.resume(123456789)
#   ```
#   """
#   @spec resume(Guild.id()) :: :ok | {:error, String.t()}
#   def resume(guild_id) do
#     voice = get_voice(guild_id)

#     cond do
#       not VoiceState.ready_for_rtp?(voice) ->
#         {:error, "Must be connected to voice channel to resume audio."}

#       VoiceState.playing?(voice) ->
#         {:error, "Audio already playing in voice channel."}

#       is_nil(voice.ffmpeg_proc) and is_nil(voice.raw_audio) ->
#         {:error, "Audio must be paused to resume."}

#       true ->
#         set_speaking(voice, true)
#         {:ok, pid} = Task.start(fn -> Audio.player_loop(voice) end)
#         update_voice(guild_id, player_pid: pid)
#         :ok
#     end
#   end

#   @doc """
#   Checks if the bot is playing sound in a voice channel.

#   ## Parameters
#     - `guild_id` - ID of guild to check if audio being played.

#   Returns `true` if the bot is currently being played in a voice channel, otherwise `false`.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.play(123456789, "https://a-real-site.biz/RickRoll.m4a")

#   iex> Remedy.Voice.playing?(123456789)
#   true

#   iex> Remedy.Voice.pause(123456789)

#   iex> Remedy.Voice.playing?(123456789)
#   false
#   ```
#   """
#   @spec playing?(Guild.id()) :: boolean
#   def playing?(guild_id) do
#     get_voice(guild_id) |> VoiceState.playing?()
#   end

#   @doc """
#   Checks if the connection is up and ready to play audio.

#   ## Parameters
#     - `guild_id` - ID of guild to check if voice connection is up.

#   Returns `true` if the bot is connected to a voice channel, otherwise `false`.

#   This function does not check if audio is already playing. For that, use `playing?/1`.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.ready?(123456789)
#   true

#   iex> Remedy.Voice.leave_channel(123456789)

#   iex> Remedy.Voice.ready?(123456789)
#   false
#   ```
#   """
#   @spec ready?(Guild.id()) :: boolean
#   def ready?(guild_id) do
#     get_voice(guild_id) |> VoiceState.ready_for_rtp?()
#   end

#   @doc """
#   Gets the id of the voice channel that the bot is connected to.

#   ## Parameters
#     - `guild_id` - ID of guild that the resultant channel belongs to.

#   Returns the `channel_id` for the channel the bot is connected to, otherwise `nil`.

#   ## Examples

#   ```Elixir
#   iex> Remedy.Voice.join_channel(123456789, 420691337)

#   iex> Remedy.Voice.get_channel(123456789)
#   420691337

#   iex> Remedy.Voice.leave_channel(123456789)

#   iex> Remedy.Voice.get_channel(123456789)
#   nil
#   ```
#   """
#   @spec get_channel_id(Guild.id()) :: Channel.id()
#   def get_channel_id(guild_id) do
#     voice = get_voice(guild_id)
#     if voice, do: voice.channel_id, else: nil
#   end

#   @doc false
#   def set_speaking(%VoiceState{} = voice, speaking, timed_out \\ false) do
#     Session.set_speaking(voice.session_pid, speaking, timed_out)
#   end

#   @doc """
#   Low-level. Set speaking flag in voice channel.

#   This function does not need to be called unless you are sending audio frames
#   directly using `Remedy.Voice.send_frames/2`.
#   """
#   @doc since: "0.5.0"
#   @spec set_is_speaking(Guild.id(), boolean) :: :ok
#   def set_is_speaking(guild_id, speaking), do: get_voice(guild_id) |> set_speaking(speaking)

#   @doc """
#   Low-level. Send pre-encoded audio packets directly.

#   Speaking should be set to true via `Remedy.Voice.set_is_speaking/2` before sending frames.

#   Opus frames will be encrypted and prefixed with the appropriate RTP header and sent immediately.
#   The length of `frames` depends on how often you wish to send a sequence of frames.
#   A single frame contains 20ms of audio. Sending more than 50 frames (1 second of audio)
#   in a single function call may result in inconsistent playback rates.

#   `Remedy.Voice.playing?/1` will not return accurate values when using `send_frames/2`
#   instead of `Remedy.Voice.play/4`
#   """
#   @doc since: "0.5.0"
#   @spec send_frames(Guild.id(), [binary]) :: :ok | {:error, String.t()}
#   def send_frames(guild_id, frames) when is_list(frames) do
#     voice = get_voice(guild_id)

#     if VoiceState.ready_for_rtp?(voice) do
#       Audio.send_frames(frames, voice)
#       :ok
#     else
#       {:error, "Must be connected to voice channel to send frames."}
#     end
#   end

#   @doc """
#   Low-level. Manually connect to voice WSStates gateway.

#   This function should only be called if config option `:voice_auto_connect` is set to `false`.
#   By default Remedy will automatically create a voice gateway when joining a channel.
#   """
#   @doc since: "0.5.0"
#   @spec connect_to_gateway(Guild.id()) :: :ok | {:error, String.t()}
#   def connect_to_gateway(guild_id) do
#     voice = get_voice(guild_id)

#     cond do
#       VoiceState.ready_for_ws?(voice) ->
#         VoiceSupervisor.create_session(voice)
#         :ok

#       is_nil(voice) ->
#         {:error, "Must be in voice channel to connect to gateway."}

#       true ->
#         {:error, "Voice gateway connection already open."}
#     end
#   end

#   @doc """
#   Listen for incoming voice RTP packets.

#   ## Parameters
#     - `guild_id` - ID of guild that the bot is listening to.
#     - `num_packets` - Number of packets to wait for.

#   Returns a list of 2-element tuples in the form `{rtp_header, opus_frame}`.

#   This function will block until the specified number of packets is received.
#   """
#   @doc since: "0.5.0"
#   @spec listen(Guild.id(), pos_integer) :: [{binary, binary}] | {:error, String.t()}
#   def listen(guild_id, num_packets) do
#     voice = get_voice(guild_id)

#     if VoiceState.ready_for_rtp?(voice) do
#       Enum.map(1..num_packets, fn _ -> Audio.get_rtp_packet(voice) end)
#     else
#       {:error, "Must be connected to voice channel to listen for incoming data."}
#     end
#   end

#   @doc false
#   def handle_call({:update, guild_id, args}, _from, state) do
#     voice =
#       state
#       |> Map.get(guild_id, VoiceState.new(guild_id: guild_id))
#       |> Map.merge(Enum.into(args, %{}))

#     state = Map.put(state, guild_id, voice)

#     if Application.get_env(:remedy, :voice_auto_connect, true),
#       do: start_if_ready(voice)

#     {:reply, voice, state}
#   end

#   @doc false
#   def handle_call({:get, guild_id}, _from, state) do
#     {:reply, Map.get(state, guild_id), state}
#   end

#   @doc false
#   def handle_call({:remove, guild_id}, _from, state) do
#     state[guild_id] |> VoiceState.cleanup()
#     VoiceSupervisor.end_session(guild_id)

#     {:reply, true, Map.delete(state, guild_id)}
#   end

#   @doc false
#   def start_if_ready(%VoiceState{} = voice) do
#     if VoiceState.ready_for_ws?(voice) do
#       VoiceSupervisor.create_session(voice)
#     end
#   end

# def update_voice_state(guild_id, channel_id, self_mute, self_deaf) do
#   case GuildShard.get_shard(guild_id) do
#     {:ok, shard_id} ->
#       SessionSupervisor
#       |> Supervisor.which_children()
#       |> Enum.filter(fn {_id, _pid, _type, [modules]} -> modules == Remedy.Shard end)
#       |> Enum.filter(fn {id, _pid, _type, _modules} -> id == shard_id end)
#       |> Enum.map(fn {_id, pid, _type, _modules} -> Supervisor.which_children(pid) end)
#       |> List.flatten()
#       |> Enum.filter(fn {_id, _pid, _type, [modules]} -> modules == Remedy.Shard.Session end)
#       |> List.first()
#       |> elem(1)
#       |> Session.update_voice_state(guild_id, channel_id, self_mute, self_deaf)

#     {:error, :id_not_found} ->
#       raise CacheError, key: guild_id, cache_name: GuildShardMapping
#   end
# end
# end
